
; Centaur Bitops Library
; Copyright (C) 2010-2011 Centaur Technology
;
; Contact:
;   Centaur Technology Formal Verification Group
;   7600-C N. Capital of Texas Highway, Suite 300, Austin, TX 78731, USA.
;   http://www.centtech.com/
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version.  This program is distributed in the hope that it will be useful but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
; more details.  You should have received a copy of the GNU General Public
; License along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.

; contextual rewriting for bitops

;; We treat loghead as a fixing function for congruence-style rewriting via
;; context-rw.lisp.  We also introduce logsquash (name subject to change),
;; which is kind of the "opposite" of loghead, but not the same as logtail.
;; (logsquash n x) is the same as (ash (logtail n x) n) or as
;; (logand (lognot (logmask n)) x), i.e. it squashes the first N bits of X to
;; 0 (whereas loghead squashes all but the first N bits of X.)

;; We also go even further with this and use logand as a fixing function when
;; its left argument is a constant.

;; principle: rewrite rules should propagate fixing functions outward,
;; and our context-rw meta rules will propagate them inward.  For example,
;; (defthmd ash-of-loghead
;;   (equal (ash (loghead n x) m)
;;          (loghead (+ (nfix n) (ifix m)) (ash x m))))
;; is a reasonable rewrite rule, because it pulls a loghead out of an ash.
;; Correspondingly,
;; (def-context-rule loghead-of-ash-context
;;   (equal (loghead n (ash (loghead (- (nfix n) (ifix m)) a) m))
;;          (loghead n (ash a m)))
;;   :fnname loghead)
;; is a context rule that propagates the loghead context into the argument of
;; ash.  This way, generally, contextual information is normalized to the
;; outside but is still available for simplifications deep inside nestings of
;; bit operations.
;; --------------------------------------------------------------------
;; Context-rw.lisp (like coi/nary) allows us to do congruence-like reasoning
;; with parameters, i.e. 


(in-package "ACL2")
(include-book "tools/rulesets" :dir :system)
(include-book "centaur/misc/context-rw" :dir :system)
(include-book "ihs/basic-definitions" :dir :system)
(include-book "ihsext-basics")
(local (include-book "ihs-extensions"))
(local (include-book "arithmetic/top-with-meta" :dir :system))
(local (include-book "equal-by-logbitp"))
(defmacro def-context-rule (name body &rest args)
  (let* ((fnname-look (ec-call (assoc-keyword :fnname args)))
         (fnname (cadr fnname-look))
         (rest-args (append (take (- (len args) (len fnname-look)) args)
                            (cddr fnname-look))))
    `(progn (defthmd ,name ,body . ,rest-args)
            (add-context-rule ,fnname (:rewrite ,name)))))



;; Since we're normalizing to logsquash of loghead, propagate a logsquash
;; context inside loghead:
(def-context-rule logsquash-of-loghead-context
  (equal (logsquash n (loghead m (logsquash n x)))
         (logsquash n (loghead m x)))
  :fnname logsquash$inline)

;; ;; Logbitp induces both a logsquash and loghead context.
;; (defthm logbitp-remove-logsquash
;;   (implies (<= (nfix m) (nfix n))
;;            (equal (logbitp n (logsquash m i))
;;                   (logbitp n i))))

;; (defthm logbitp-remove-loghead
;;   (implies (< (nfix n) (nfix m))
;;            (equal (logbitp n (loghead m i))
;;                   (logbitp n i))))

(def-context-rule logbitp-induces-logsquash/loghead-context
  (implies (syntaxp (not (quotep n)))
           ;; if n is quotep we'll prefer a logand context instead
           (equal (logbitp n (logsquash n (loghead (+ 1 (nfix n)) i)))
                  (logbitp n i)))
  :fnname logbitp)

;; ;; Logtail induces a logsquash context.
;; (defthm logtail-remove-logsquash
;;   (implies (<= (nfix m) (nfix n))
;;            (equal (logtail n (logsquash m i))
;;                   (logtail n i))))

(def-context-rule logtail-induces-logsquash-context
  (equal (logtail n (logsquash n i))
         (logtail n i))
  :fnname logtail$inline)

;; It also passes a (modified) loghead context.
;; We already have logtail-of-loghead to remove this context.
(def-context-rule logtail-pass-loghead-context
  (equal (loghead n (logtail m (loghead (+ (nfix n) (nfix m)) i)))
         (loghead n (logtail m i)))
  :hints(("Goal" :in-theory (disable logsquash)))
  :fnname loghead$inline)


;; Logic ops are transparent to both types of context.

;; Logior:
(defthm logior-remove-loghead-1
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logior (loghead m a) b))
                  (loghead n (logior a b)))))

(defthm logior-remove-loghead-2
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logior a (loghead m b)))
                  (loghead n (logior a b)))))

(def-context-rule logior-pass-loghead-context-1
  (equal (loghead n (logior (loghead n a) b))
         (loghead n (logior a b)))
  :fnname loghead$inline)

(def-context-rule logior-pass-loghead-context-2
  (equal (loghead n (logior a (loghead n b)))
         (loghead n (logior a b)))
  :fnname loghead$inline)

(defthm logior-remove-logsquash-1
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logior (logsquash m a) b))
                  (logsquash n (logior a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(defthm logior-remove-logsquash-2
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logior a (logsquash m b)))
                  (logsquash n (logior a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(def-context-rule logior-pass-logsquash-context-1
  (equal (logsquash n (logior (logsquash n a) b))
         (logsquash n (logior a b)))
  :fnname logsquash$inline)

(def-context-rule logior-pass-logsquash-context-2
  (equal (logsquash n (logior a (logsquash n b)))
         (logsquash n (logior a b)))
  :fnname logsquash$inline)

;; Logand:
(defthm logand-remove-loghead-1
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logand (loghead m a) b))
                  (loghead n (logand a b)))))

(defthm logand-remove-loghead-2
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logand a (loghead m b)))
                  (loghead n (logand a b)))))

(def-context-rule logand-pass-loghead-context-1
  (equal (loghead n (logand (loghead n a) b))
         (loghead n (logand a b)))
  :fnname loghead$inline)

(def-context-rule logand-pass-loghead-context-2
  (equal (loghead n (logand a (loghead n b)))
         (loghead n (logand a b)))
  :fnname loghead$inline)

(defthm logand-remove-logsquash-1
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logand (logsquash m a) b))
                  (logsquash n (logand a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(defthm logand-remove-logsquash-2
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logand a (logsquash m b)))
                  (logsquash n (logand a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(def-context-rule logand-pass-logsquash-context-1
  (equal (logsquash n (logand (logsquash n a) b))
         (logsquash n (logand a b)))
  :fnname logsquash$inline)

(def-context-rule logand-pass-logsquash-context-2
  (equal (logsquash n (logand a (logsquash n b)))
         (logsquash n (logand a b)))
  :fnname logsquash$inline)

;; Logxor:
(defthm logxor-remove-loghead-1
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logxor (loghead m a) b))
                  (loghead n (logxor a b)))))

(defthm logxor-remove-loghead-2
  (implies (<= (nfix n) (nfix m))
           (equal (loghead n (logxor a (loghead m b)))
                  (loghead n (logxor a b)))))

(def-context-rule logxor-pass-loghead-context-1
  (equal (loghead n (logxor (loghead n a) b))
         (loghead n (logxor a b)))
  :fnname loghead$inline)

(def-context-rule logxor-pass-loghead-context-2
  (equal (loghead n (logxor a (loghead n b)))
         (loghead n (logxor a b)))
  :fnname loghead$inline)

(defthm logxor-remove-logsquash-1
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logxor (logsquash m a) b))
                  (logsquash n (logxor a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(defthm logxor-remove-logsquash-2
  (implies (<= (nfix m) (nfix n))
           (equal (logsquash n (logxor a (logsquash m b)))
                  (logsquash n (logxor a b))))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs))))

(def-context-rule logxor-pass-logsquash-context-1
  (equal (logsquash n (logxor (logsquash n a) b))
         (logsquash n (logxor a b)))
  :fnname logsquash$inline)

(def-context-rule logxor-pass-logsquash-context-2
  (equal (logsquash n (logxor a (logsquash n b)))
         (logsquash n (logxor a b)))
  :fnname logsquash$inline)

;; Lognot:
;; (defthm lognot-remove-loghead
;;   (implies (<= (nfix n) (nfix m))
;;            (equal (loghead n (lognot (loghead m a)))
;;                   (loghead n (lognot a)))))

(def-context-rule lognot-pass-loghead-context
  (equal (loghead n (lognot (loghead n a)))
         (loghead n (lognot a)))
  :fnname loghead$inline)

;; (defthm lognot-remove-logsquash
;;   (implies (<= (nfix m) (nfix n))
;;            (equal (logsquash n (lognot (logsquash m a)))
;;                   (logsquash n (lognot a)))))

(def-context-rule lognot-pass-logsquash-context
  (equal (logsquash n (lognot (logsquash n a)))
         (logsquash n (lognot a)))
  :fnname logsquash$inline)



;; Ash propagates a modified loghead/logsquash context.  Right shift also
;; induces a logsquash context (same as logtail-induces-logsquash-context).

;; This is not compatible with loghead-of-ash.
(defthmd ash-of-loghead
  (equal (ash (loghead m i) n)
         (loghead (+ (nfix m) (ifix n)) (ash i n)))
  :hints(("Goal" :in-theory (e/d (loghead-of-ash ifix nfix loghead**)
                                 (loghead-identity)))))

(local (in-theory (disable loghead-of-ash)))

(def-context-rule ash-propagate-loghead-context
  (equal (loghead n (ash (loghead (- (nfix n) (ifix m)) i) m))
         (loghead n (ash i m)))
  :hints(("Goal" :in-theory (e/d (ash-of-loghead
                                  ifix nfix loghead-of-loghead-split
                                  loghead**)
                                 (loghead-identity))))
  :fnname loghead$inline)

(defthmd ash-of-logsquash
  (equal (ash (logsquash m i) n)
         (logsquash (+ (nfix m) (ifix n)) (ash i n)))
  :hints(("Goal" :in-theory (e/d (logsquash-of-ash ifix nfix
                                  logsquash**)))))

(def-context-rule ash-propagate-logsquash-context
  (equal (logsquash n (ash (logsquash (- (nfix n) (ifix m)) i) m))
         (logsquash n (ash i m)))
  :hints(("Goal" :in-theory (enable logsquash-of-ash)))
  :fnname logsquash$inline)

(def-context-rule right-shift-induces-logsquash-context
  (implies (<= (ifix m) 0)
           (equal (ash (logsquash (- (ifix m)) i) m)
                  (ash i m)))
  :hints(("Goal" :in-theory (e/d ()
                                 (right-shift-to-logtail))
          :use ((:instance logsquash-of-ash
                 (x i) (n 0)))))
  :fnname ash)
  


;; Addition/subtraction pass loghead, but not logsquash, contexts

(local
 (encapsulate
   nil
   (local (defun ind (n m a b c)
            (if (zp n)
                (list m a b c)
              (ind (1- n) (1- m)
                   (logcdr a)
                   (logcdr b)
                   (b-ior (b-and (logcar a) (logcar b))
                          (b-ior
                           (b-and (logcar a) (logcar c))
                           (b-and (logcar b) (logcar c))))))))

; Reordering the rewrite-clause-type-alist: I added the uppercase text below to
; make this work.  See the comment in rewrite-clause-type-alist.
; JSM April 7, 2013.

   (defthm loghead-of-sum-lemma
     (implies (and (integerp a)
                   (integerp b)
                   (bitp c)
                   (<= (nfix n) (nfix m)))
              (equal (loghead n (+ c a (loghead m b)))
                     (loghead n (+ c a b))))
     :hints (("goal"
              :induct (ind n m a b c)
              :do-not '(generalize fertilize)
              :expand ((:free (b) (loghead n b))
                       (:free (b) (loghead m b)))
              :in-theory (e/d (nfix logcdr-of-bit b-and b-ior)
                              (loghead-identity
                               (:FORWARD-CHAINING LOGCAR-POSSIBILITIES)
                               ))
              :do-not-induct t))
     :rule-classes nil)))


;; Rewrite rules for plus/minus under loghead
(defthmd loghead-of-minus-of-loghead
  (implies (and (integerp b)
                (<= (nfix n) (nfix m)))
           (equal (loghead n (- (loghead m b)))
                  (loghead n (- b))))
  :hints (("goal" :in-theory (e/d (minus-to-lognot))
           :use ((:instance loghead-of-sum-lemma
                  (c 1) (a 0) (b (lognot (loghead m b))))
                 (:instance loghead-of-sum-lemma
                  (c 1) (a 0) (b (lognot b)))))))

(defthmd loghead-of-plus-loghead-first
  (implies (and (integerp a)
                (integerp b)
                (<= (nfix n) (nfix m)))
           (equal (loghead n (+ (loghead m b) a))
                  (loghead n (+ a b))))
  :hints (("goal" :use ((:instance loghead-of-sum-lemma (c 0))))))

(defthmd loghead-of-plus-loghead-second
  (implies (and (integerp a)
                (integerp b)
                (<= (nfix n) (nfix m)))
           (equal (loghead n (+ a (loghead m b)))
                  (loghead n (+ a b))))
  :hints (("goal" :use ((:instance loghead-of-sum-lemma (c 0))))))

;; Context rules for plus/minus
(def-context-rule loghead-of-minus-context
  (implies (integerp b)
           (equal (loghead n (- (loghead n b)))
                  (loghead n (- b))))
  :hints(("Goal" :in-theory (enable loghead-of-minus-of-loghead)))
  :fnname loghead$inline)

(def-context-rule loghead-of-plus-context-1
  (implies (and (integerp a) (integerp b))
           (equal (loghead n (+ (loghead n a) b))
                  (loghead n (+ a b))))
  :hints(("Goal" :in-theory (enable loghead-of-plus-loghead-second)))
  :fnname loghead$inline)

(def-context-rule loghead-of-plus-context-2
  (implies (and (integerp a) (integerp b))
           (equal (loghead n (+ a (loghead n b)))
                  (loghead n (+ a b))))
  :hints(("Goal" :in-theory (enable loghead-of-plus-loghead-second)))
  :fnname loghead$inline)





;; Logand.  This is the most general, but only works with a constant as the
;; mask.  Doesn't work with add/subtract.

(def-context-rule logbitp-induces-logand-context
  (implies (syntaxp (quotep n))
           (equal (logbitp n (logand (ash 1 (nfix n)) m))
                  (logbitp n m)))
  :fnname logbitp)

(def-context-rule logior-induces-logand-context
  (implies (syntaxp (quotep n))
           (equal (logior n (logand (lognot n) m))
                  (logior n m)))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logior)

(def-context-rule lognot-pass-logand-context
  (implies (syntaxp (quotep n))
           (equal (logand n (lognot (logand n m)))
                  (logand n (lognot m))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logior-pass-logand-context-1
  (implies (syntaxp (quotep n))
           (equal (logand n (logior (logand n m1) m2))
                  (logand n (logior m1 m2))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logior-pass-logand-context-2
  (implies (syntaxp (quotep n))
           (equal (logand n (logior m2 (logand n m1)))
                  (logand n (logior m2 m1))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logand-pass-logand-context-1
  (implies (syntaxp (quotep n))
           (equal (logand n (logand (logand n m1) m2))
                  (logand n (logand m1 m2))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logand-pass-logand-context-2
  (implies (syntaxp (quotep n))
           (equal (logand n (logand m2 (logand n m1)))
                  (logand n (logand m2 m1))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logxor-pass-logand-context-1
  (implies (syntaxp (quotep n))
           (equal (logand n (logxor (logand n m1) m2))
                  (logand n (logxor m1 m2))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

(def-context-rule logxor-pass-logand-context-2
  (implies (syntaxp (quotep n))
           (equal (logand n (logxor m2 (logand n m1)))
                  (logand n (logxor m2 m1))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)


(defthm logand-loghead-combine-contexts
  (implies (syntaxp (and (quotep n1)
                         (quotep n2)))
           (equal (logand n1 (loghead n2 m))
                  (logand (logand n1 (1- (ash 1 (nfix n2)))) m)))
  :hints (("goal" :use ((:instance logand-with-bitmask
                         (mask (1- (ash 1 (nfix n2))))
                         (i m)))
           :in-theory (e/d ()
                           (ash-1-removal)))))

(defthm loghead-logand-combine-contexts
  (implies (syntaxp (and (quotep n1)
                         (quotep n2)))
           (equal (loghead n1 (logand n2 m))
                  (logand (logand n2 (1- (ash 1 (nfix n1)))) m)))
  :hints (("goal" :use ((:instance logand-with-bitmask
                         (mask (1- (ash 1 (nfix n1))))
                         (i (logand n2 m))))
           :in-theory (disable ash-1-removal))
          (equal-by-logbitp-hammer)))

(defthm logand-logsquash-combine-contexts
  (implies (syntaxp (and (quotep n1)
                         (quotep n2)))
           (equal (logand n1 (logsquash n2 m))
                  (logand (logand n1 (lognot (1- (ash 1 (nfix n2))))) m)))
  :hints (("goal" :in-theory (disable ash-1-removal))
          (equal-by-logbitp-hammer)))

(defthm logsquash-logand-combine-contexts
  (implies (syntaxp (and (quotep n1)
                         (quotep n2)))
           (equal (logsquash n1 (logand n2 m))
                  (logand (logand n2 (lognot (1- (ash 1 (nfix n1))))) m)))
  :hints (("goal" :use ((:instance logand-with-negated-bitmask
                         (mask (lognot (1- (ash 1 (nfix n1)))))
                         (i (logand n2 m))))
           :in-theory (e/d () (ash-1-removal
                               logand-with-negated-bitmask)))))


(def-context-rule ash-propagate-logand-context
  (implies (syntaxp (quotep n))
           (equal (logand n (ash (logand (ash n (- (ifix m))) i) m))
                  (logand n (ash i m))))
  :hints ((equal-by-logbitp-hammer))
  :fnname binary-logand)

;; Rewrite a logsquash of loghead to a logand when sizes are constant:
(defthm logsquash-of-loghead-to-logand
  (implies (syntaxp (and (quotep n) (quotep m)))
           (equal (logsquash n (loghead m i))
                  (logand (logsquash n (loghead m -1)) i)))
  :hints(("Goal" :in-theory (enable* ihsext-inductions
                                     ihsext-recursive-redefs)
          :induct (and (logsquash n i)
                       (loghead m i)))))

(local (defthm logbitp-when-gte-integer-length
         (implies (<= (integer-length n) (nfix i))
                  (equal (logbitp i n)
                         (< (ifix n) 0)))
         :hints (("goal" :in-theory (enable* ihsext-inductions
                                             ihsext-recursive-redefs)
                  :induct (logbitp i n)
                  :expand ((integer-length n))))))

;; logand can create a loghead context, but for the most part we don't want it
;; to:
(defthmd logand-of-loghead-context-lemma
  (implies (<= 0 (ifix n))
           (equal (logand n (loghead (integer-length n) b))
                  (logand n b)))
  :hints ((equal-by-logbitp-hammer)))

;; Logand can create a loghead context for +/-
;; Context rules for plus/minus
(def-context-rule logand-of-minus-context
  (implies (and (syntaxp (quotep n))
                (<= 0 (ifix n))
                (integerp b))
           (equal (logand n (- (loghead (integer-length n) b)))
                  (logand n (- b))))
  :hints(("Goal" :use ((:instance logand-of-loghead-context-lemma
                        (b (- b)))
                       (:instance logand-of-loghead-context-lemma
                        (b (- (loghead (integer-length n) b)))))
          :in-theory (enable loghead-of-minus-of-loghead)))
  :fnname binary-logand)

(def-context-rule logand-of-plus-context-first
  (implies (and (syntaxp (quotep n))
                (<= 0 (ifix n))
                (integerp a)
                (integerp b))
           (equal (logand n (+ (loghead (integer-length n) a) b))
                  (logand n (+ a b))))
  :hints(("Goal" :use ((:instance logand-of-loghead-context-lemma
                        (b (+ a b)))
                       (:instance logand-of-loghead-context-lemma
                        (b (+ (loghead (integer-length n) a) b))))
          :in-theory (enable loghead-of-plus-loghead-first
                             loghead-of-plus-loghead-second)))
  :fnname binary-logand)

(def-context-rule logand-of-plus-context-second
  (implies (and (syntaxp (quotep n))
                (<= 0 (ifix n))
                (integerp a)
                (integerp b))
           (equal (logand n (+ a (loghead (integer-length n) b)))
                  (logand n (+ a b))))
  :hints(("Goal" :use ((:instance logand-of-loghead-context-lemma
                        (b (+ a b)))
                       (:instance logand-of-loghead-context-lemma
                        (b (+ a (loghead (integer-length n) b)))))
          :in-theory (enable loghead-of-plus-loghead-first
                             loghead-of-plus-loghead-second)))
  :fnname binary-logand)

(def-ruleset! bitops-congruences
  '(apply-context-for-loghead$inline
    apply-context-for-logsquash$inline
    apply-context-for-binary-logior
    apply-context-for-binary-logand
    common-lisp::apply-context-for-logbitp
    common-lisp::apply-context-for-ash
    apply-context-for-logtail$inline
    loghead-of-logsquash-commute
    logsquash-of-loghead-zero
    logsquash-idempotent
    logsquash-of-logsquash-split
    loghead-of-loghead-split
    logbitp-of-logsquash-in-bounds
    logbitp-of-loghead-out-of-bounds
    logior-remove-loghead-1
    logior-remove-loghead-2
    logior-pass-loghead-context-1
    logior-pass-loghead-context-2
    logior-remove-logsquash-1
    logior-remove-logsquash-2
    logior-pass-logsquash-context-1
    logior-pass-logsquash-context-2
    logand-remove-loghead-1
    logand-remove-loghead-2
    logand-pass-loghead-context-1
    logand-pass-loghead-context-2
    logand-remove-logsquash-1
    logand-remove-logsquash-2
    logand-pass-logsquash-context-1
    logand-pass-logsquash-context-2
    logxor-remove-loghead-1
    logxor-remove-loghead-2
    logxor-pass-loghead-context-1
    logxor-pass-loghead-context-2
    logxor-remove-logsquash-1
    logxor-remove-logsquash-2
    logxor-pass-logsquash-context-1
    logxor-pass-logsquash-context-2
    loghead-cancel-in-lognot
    logsquash-cancel-in-lognot
    lognot-pass-loghead-context
    lognot-pass-logsquash-context
    ash-of-loghead
    ash-of-logsquash
    loghead-of-minus-of-loghead
    loghead-of-plus-loghead-first
    loghead-of-plus-loghead-second
    loghead-of-minus-context
    loghead-of-plus-context-1
    loghead-of-plus-context-2
    logbitp-induces-logand-context
    logior-induces-logand-context
    lognot-pass-logand-context
    logior-pass-logand-context-1
    logior-pass-logand-context-2
    logand-pass-logand-context-1
    logand-pass-logand-context-2
    logxor-pass-logand-context-1
    logxor-pass-logand-context-2
    logand-loghead-combine-contexts
    loghead-logand-combine-contexts
    logand-logsquash-combine-contexts
    logsquash-logand-combine-contexts
    ash-propagate-logand-context
    logsquash-of-loghead-to-logand
    logand-of-minus-context
    logand-of-plus-context-first
    logand-of-plus-context-second))

(def-ruleset! bitops-congruence-incompatible
  '(;; loghead-of-logior
    loghead-of-ash
    logsquash-of-ash
    ;; loghead-of-logxor
    ;; loghead-of-logand
    ))
;; logand-with-bitmask?

(in-theory (disable* apply-context-for-loghead$inline
                     apply-context-for-logsquash$inline
                     apply-context-for-binary-logior
                     apply-context-for-binary-logand
                     common-lisp::apply-context-for-logbitp
                     common-lisp::apply-context-for-ash
                     apply-context-for-logtail$inline))
